from opcodetools.assembler.assembler import Assembler, ASMException

import sys
import os

# TODO: lots of command line options:
#  - Listing file (yes/no)
#  - Output file (name)

try:
    a = Assembler(sys.argv[1])
    a.assemble()
    a.write_binary(sys.argv[1] + '.bin')
    a.write_listing(sys.argv[1] + '.lst')
except ASMException as ex:
    print('##', str(ex))
    print('##', ex.line)



class Opcode():

    '''All the information about a single opcode

    Opcodes objects are super-simple. They are just structures.
    '''

    def __init__(self, info: dict):
        self.cpu = None
        self.frags = None  # For assembly
        self.info = info
        self.mnemonic = info['mnemonic']
        self.code = []
        code = info['code']
        cts = {}  # Byte sizes for each spec
        for i in range(0, len(code), 2):
            frag = code[i:i + 2]
            if frag.islower():
                self.code.append(frag)
                if not frag[0] in cts:
                    cts[frag[0]] = 1
                else:
                    cts[frag[0]] += 1
            else:
                self.code.append(int(frag, 16))
        self.use = {}
        use = info['use']
        if use:
            use = use.split(',')
            for u in use:
                i = u.index('=')
                sp = u[:i].strip()
                specs = u[i + 1:].strip().split('_')
                specs.append('s' + str(cts[sp]))
                self.use[sp] = specs



import opcodetools.cpu.base_assembly
import opcodetools.cpu.base_disassembly
import opcodetools.cpu.opcode

'''
  The CPU class is a list of individual opcodes. An opcode has several parts:
    - DATA the byte format of the opcode in memory
    - MNEMONIC the assembly language representation
    - USE how the individual parts (lower case letters) of the opcode bytes are used

  Example opcodes from different machines:

    DATA        MMNEMONIC     CPU   USE
    20 a0 a1    JSR a         6502  (a=code)
    90 aa       BCC a         6502  (a=code_pcr)
    21 aa       AND (a,X)     6502  (a=ram)
    69 aa       ADC #a        6502  (a=const)
    D3 aa       OUT (a),A     Z80   (a=port_w)
    6C a0 a1    JMP (a)       6502  (a=code)
    10 aa bb    JBC a,b       8502  (a=ram_bit,b=code)
    85 bb aa    MOV a,b       6502  (a=ram,b=ram)

  Lower-case letters identify substitution information in the opcode. The single letters in mnemonic map to
  multiple letters in the data list. The USE field explains how each substitution is used (see below).

  Multibyte substitutions list the order of the bytes. 0 is the least significant, 1 is next, and so on.
  A little-endian address would be: a0a1. A big-endian address would be: a1a0.

  The USE field explains how the field is used. The webifier uses this information to pick label links for
  the webified disassembly.

  The base uses are:
    - const (an immediate constant)
    - data  (address is a regular read/write memory access)
    - code  (address is a label in the code)
    - port  (address is a port read/write memory access)

  Direction qualifiers include (direction often matters for I/O addresses):
    - _r  The opcode reads from the address
    - _w  The opcode writes to the address
    - _rw The opcode reads from the address and then writes back to it

  The _pcr qualifier indicates the actual address is relative to the program counter.

  Some processors have their own substitution qualifiers.

  Qualifiers - 6809 specific
    _pair  Pair of registers
    _pshs  Register list (push to S)
    _puls  Register list (pull from S)
    _pshu  Register list (push to U)
    _pulu  Register list (pull from U)
    _bp    BasePage *

    * The 6809 has a BP register (not always 00). Thus address 0010 and address >10 could be different.
    Other processors have shortcuts for 1 byte addresses, but the upper byte is always 00 such
    that 0010 and >10 are the same.

  Qualifiers - 8051 specific
    _bit  Bit address
    _11   11_bit address (3 bits in opcode)

'''


class CPUException(Exception):
    pass


class CPU(opcodetools.cpu.base_assembly.BaseAssembly, opcodetools.cpu.base_disassembly.BaseDisassembly):

    cpu_singletons = {}

    def __init__(self, opcodes: list, little_endian=True):
        '''Create a basic CPU from a list of opcodes

        Args:
            opcodes (list): List of opcode objects
            little_endian (bool): True for little-endian machines (False for big-endian)
        '''

        self.little_endian = little_endian
        self._opcodes = []
        self._opcodes_info = opcodes
        self._quick_codes = {}  # Helps in searching for large opcode lists
        for op in opcodes:
            opc = self.make_opcode(op)
            opc._cpu = self
            self._opcodes.append(opc)
            key = opc.code[0]
            if not isinstance(key, int):
                raise CPUException('First code byte must be a number')
            if key in self._quick_codes:
                self._quick_codes[key].append(opc)
            else:
                self._quick_codes[key] = [opc]

    def make_opcode(self, info: dict) -> opcodetools.cpu.opcode.Opcode:
        '''Create an opcode from the given info.

        This is a factory method to allow for overrides

        Args:
            info (dict): the opcode information
        Returns:
            Opcode: factoried opcode
        '''
        return opcodetools.cpu.opcode.Opcode(info)

    def make_word(self, term: int) -> int:
        '''Convert an integer to a list of 2 bytes

        Args:
            term (int): the value to convert
        Returns:
            list: two byte-values in the CPU's endian order
        '''

        a = term & 0xFF       # Little
        b = (term >> 8) & 0xFF  # Big

        if self.little_endian:
            return [a, b]
        else:
            return [b, a]



import setuptools

# https://packaging.python.org/tutorials/packaging-projects/

# From the directory with this file
#   (modify the version then ...)
#   py setup.py sdist bdist_wheel
#   py -m twine upload dist/*

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="opcodetools",  # Replace with your own username
    version="1.5",
    author="Chris Cantrell",
    author_email="topherCantrell@gmail.com",
    description="Assemblers/Disassemblers for retro processors (Z80, 6502, 6809, etc)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/topherCantrell/opcodetools",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



import opcodetools.cpu.base_cpu


_CPUS = {}


def get_cpu_by_name(name: str) -> opcodetools.cpu.base_cpu.CPU:

    if name in _CPUS:
        return _CPUS[name]

    # Delayed imports ... only for the processors we need

    cp = None
    if name == '6502':
        import opcodetools.cpu.cpu_6502
        cp = opcodetools.cpu.cpu_6502.CPU_6502()
    elif name == '6803':
        import opcodetools.cpu.cpu_6803
        cp = opcodetools.cpu.cpu_6803.CPU_6803()
    elif name == '6809':
        import opcodetools.cpu.cpu_6809
        cp = opcodetools.cpu.cpu_6809.CPU_6809()
    elif name == '8052':
        import opcodetools.cpu.cpu_8052
        cp = opcodetools.cpu.cpu_8052.CPU_8052()
    elif name == 'DVG':
        import opcodetools.cpu.cpu_DVG
        cp = opcodetools.cpu.cpu_DVG.CPU_DVG()
    elif name == 'Z80':
        import opcodetools.cpu.cpu_Z80
        cp = opcodetools.cpu.cpu_Z80.CPU_Z80()
    elif name == 'Z80GB':
        import opcodetools.cpu.cpu_Z80GB
        cp = opcodetools.cpu.cpu_Z80GB.CPU_Z80GB()

    if cp:
        _CPUS[name] = cp

    return cp



from opcodetools.cpu.opcode import Opcode


class BaseDisassembly:

    def init_disassembly(self):
        pass

    def get_field_spacing(self):
        '''Return the field spacing for disassembly

        Specific CPUs should override this for their own spacings

        Returns:
            dict: spacing for parts of disassembly
        '''

        return {
            'address_size': 4,  # Number of digits
            'data': 16,  # Spacing reserved for the data section
            'mnemonic': [8, 20]  # Spacing for each part of the mnemonic
        }

    def binary_to_string_unknown(self, address: int, opvalue: int):
        '''Make a disassembly string for an unknown opcode

        Args:
            address (int): the address
            opvalue (int): the unknown byte value
        Returns:
            str: the string to print
        '''

        s = hex(address)[2:].upper().rjust(4, '0')
        v = hex(opvalue)[2:].upper().rjust(2, '0')
        return s + ': ' + v + ' ; ????'

    def binary_to_string(self, opcode: Opcode, binary: list, address: int, fills: dict):
        '''Make a disassembly string from the given info

        Args:
            opcode (Opcode): the opcode
            address (int): the address of the opcode
            binary (list): the opcode data
            fills (dict): the opcode fill-in information

        Returns:
            The complete disassembly string
        '''

        # Spacing for the disassembly fields
        spa = self.get_field_spacing()

        # Address
        fs = '{:0' + str(spa['address_size']) + 'X}'
        add = fs.format(address)

        # Data
        ds = ''
        for i in range(len(binary)):
            ds = ds + '{:02X} '.format(binary[i])
        ds = ds.ljust(spa['data'], ' ')

        # Multi-word mnemonic spacing
        # TODO: use however many words are in the array ... or just the one if it isn't an array
        mn = opcode.mnemonic
        i = mn.find(' ')
        if i >= 0:
            a = mn[0:i]
            b = mn[i + 1:]
            mn = a.ljust(spa['mnemonic'][0]) + b.ljust(spa['mnemonic'][1])
        else:
            mn = mn.ljust(spa['mnemonic'][0] + spa['mnemonic'][1])

        # Build the basic form
        base = f'{add}: {ds}{mn}'

        for f in fills:
            fill_info = fills[f]
            i = base.find(f)
            if i >= 0:
                # TODO: visible spacing
                base = base.replace(f, fill_info['sub_value'])

        return base

    def binary_to_string_fill(self, address: int, binary: list, opcode: Opcode, fills: dict, ind: int):
        '''Fill in an opcode data value

        Different processors might override this for their own special needs. The
        "binary_to_string" defers to this method.

        Entries are added to fills for example:
            'p' : {
                'sub_value' : '$1024', # The actual value (string)
                'visual_size' : 5      # Number of printed characters in sub (might include HTML)
                'numeric_value' : 0    # Actual numeric value
            }

        Args:
            address (int): the address of the opcode
            binary (list): the binary data for the opcode
            opcode (Opcode): the Opcode
            fills (dict): dictionary of fill-in values (add to this)
            ind (int): bytes index of the fill-in (binary and opcode)

        '''

        # TODO: clunky with the multi-byte values.

        # Find/make the fill-in entry
        spec = opcode.code[ind]
        if spec[0] not in fills:
            fills[spec[0]] = {'sub_value': '$00', 'visual_size': 3, 'numeric_value': 0}
            entry = fills[spec[0]]
        else:
            entry = fills[spec[0]]

        ov = entry['numeric_value']

        # New numeric value
        val = binary[ind]
        if spec[1] == '1':
            val = val * 256
            entry['visual_size'] = 5
        val += ov
        entry['numeric_value'] = val

        # New substitution string
        fs = '${:0' + str(entry['visual_size'] - 1) + 'X}'

        if 'pcr' in opcode.use[spec[0]]:
            if 's1' in opcode.use[spec[0]]:
                # One byte relative (from start of next instruction)
                fa = address + len(opcode.code)
                if val < 0x80:
                    fa = fa + val
                else:
                    fa = fa + (val - 256)
            else:
                # TODO: support for more than s1,s2
                # Two byte relative (from start of next instruction)
                fa = address + len(opcode.code)
                if val < 0x8000:
                    fa = fa + val
                else:
                    fa = fa + (val - 65536)
            entry['relative_target'] = fa
            val = fa

        entry['sub_value'] = fs.format(val)

    def get_mnemonic_fills(self, opcode: Opcode, binary: list, address: int):
        '''Get all the fill-in information for a given opcode

        Args:
            opcode (Opcode): the opcode
            address (int): the address of the opcode
            binary (list): the binary data for the opcode

        Returns:
            dict: all the fill ins
        '''

        code = opcode.code
        fills = {}
        for i in range(len(binary)):
            g = code[i]
            if isinstance(g, str):
                # Call out to the specific CPU in case it has specials
                self.binary_to_string_fill(address, binary, opcode, fills, i)

        return fills

    def find_opcodes_for_binary(self, binary: list, exact: bool=False) -> list:
        '''Find the opcodes that matches the binary (a disassembly operation)

        If the match is not and exact match, the the binary can include lots of extra
        bytes (since we don't know exactly how many until we find the match).

        Args:
            binary (list[int]): the bytes to disassemble
            start (int): starting point in the bytes [default is 0]
            exact (bool): True if the opcode must match the given bytes exactly [default is False]
        Returns:
            list[Opcode]: The opcodes information
        '''

        end = len(binary)

        if not end:
            return []

        if not binary[0] in self._quick_codes:
            return []

        possible = self._quick_codes[binary[0]]

        ret = []

        for oc in possible:
            if exact and len(oc.code) != end:
                continue
            code = oc.code
            could_be = True
            for b in range(len(code)):
                if b >= end:
                    could_be = False
                    break
                if isinstance(code[b], str):
                    # Matches anything
                    continue
                if code[b] != binary[b]:
                    could_be = False
                    break
            if could_be:
                ret.append(oc)

        return ret
